LINGER = '_linger'

LINGER_MIX_INT8_MANUAL_ROUND_LAYERS = LINGER+'_round_tensor'

LINGER_DUMP_NAME = LINGER+"_dump_name"

LINGER_IGNORE_PAMAMTER = LINGER+"_ignore_parameter"

LINGER_IQTENSOR_LAYER_COUNTER = LINGER+"_iq_tensor_index_"

LINGER_FUNCINT_BMM_COUNTER = LINGER+"_funcint_bmm_index_"

LINGER_MODE = LINGER+"_mode"

LINGER_OBIT = LINGER+"_obit"

LINGER_AHEAD_RELU = LINGER+"_ahead_relu"

LINGER_AHEAD_SIGMOID = LINGER+"_ahead_sigmoid"


__all__ = ['LINGER_MIX_INT8_MANUAL_ROUND_LAYERS', 'LINGER_DUMP_NAME', 'LINGER_IGNORE_PAMAMTER', 'LINGER_IQTENSOR_LAYER_COUNTER',
           'LINGER_FUNCINT_BMM_COUNTER', 'LINGER_OBIT', 'LINGER_MODE', 'LINGER_AHEAD_RELU', 'LINGER_AHEAD_SIGMOID']
