from setuptools import find_packages, setup

setup(
    name="pylinger",
    version="1.0.0",
    description="linger is package of fix training",
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Mathematics',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Programming Language :: C++',
            'Programming Language :: Python :: 3',
    ]
)
